@@ -36,18 +36,13 @@
 #include "gnome-vfs-ops.h"
 #include "gnome-vfs-result.h"
 #include "gnome-vfs-uri.h"
-#include <dirent.h>
-#include <regex.h>
 #include <stdio.h>
 #include <string.h>
 #include <time.h>
 
-#define DEFAULT_DATE_TRACKER_INTERVAL	5	/* in milliseconds */
+#include <glib/gstdio.h>
 
-typedef struct {
-	char *mime_type;
-	regex_t regex;
-} RegexMimePair;
+#define DEFAULT_DATE_TRACKER_INTERVAL	5	/* in milliseconds */
 
 typedef struct {
 	char *file_path;
@@ -64,7 +59,7 @@ struct FileDateTracker {
 #ifdef G_THREADS_ENABLED
 
 /* We lock this mutex whenever we modify global state in this module.  */
-G_LOCK_DEFINE_STATIC (mime_mutex);
+G_LOCK_DEFINE (gnome_vfs_mime_mutex);
 
 #endif /* G_LOCK_DEFINE_STATIC */
 
@@ -72,17 +67,20 @@ G_LOCK_DEFINE_STATIC (mime_mutex);
 /**
  * gnome_vfs_mime_shutdown:
  *
+ * Deprecated: This function doesn't have to be called as the
+ * operating system automatically cleans up resources when exiting.
+ *
  * Unload the MIME database from memory.
  **/
 
 void
 gnome_vfs_mime_shutdown (void)
 {
-	G_LOCK (mime_mutex);
+	G_LOCK (gnome_vfs_mime_mutex);
 
 	xdg_mime_shutdown ();
 
-	G_UNLOCK (mime_mutex);
+	G_UNLOCK (gnome_vfs_mime_mutex);
 }
 
 /**
@@ -106,7 +104,15 @@ gnome_vfs_mime_type_from_name_or_default (const char *filename, const char *defa
 		return defaultv;
 	}
 
-	separator = g_utf8_strrchr (filename, -1, '/');
+	separator = strrchr (filename, '/');
+#ifdef G_OS_WIN32
+	{
+		const char *sep2 = strrchr (filename, '\\');
+		if (separator == NULL ||
+		    (sep2 != NULL && sep2 > separator))
+			separator = sep2;
+	}
+#endif
 	if (separator != NULL) {
 		separator++;
 		if (*separator == '\000')
@@ -115,9 +121,9 @@ gnome_vfs_mime_type_from_name_or_default (const char *filename, const char *defa
 		separator = filename;
 	}
 
-	G_LOCK (mime_mutex);
+	G_LOCK (gnome_vfs_mime_mutex);
 	mime_type = xdg_mime_get_mime_type_from_file_name (separator);
-	G_UNLOCK (mime_mutex);
+	G_UNLOCK (gnome_vfs_mime_mutex);
 
 	if (mime_type)
 		return mime_type;
@@ -167,7 +173,9 @@ _gnome_vfs_read_mime_from_buffer (GnomeVFSMimeSniffBuffer *buffer)
 	GnomeVFSResult result = GNOME_VFS_OK;
 	const char *mime_type;
 
+	G_LOCK (gnome_vfs_mime_mutex);
 	max_extents = xdg_mime_get_max_buffer_extents ();
+	G_UNLOCK (gnome_vfs_mime_mutex);
 	max_extents = CLAMP (max_extents, 0, MAX_SNIFF_BUFFER_ALLOWED);
 
 	if (!buffer->read_whole_file) {
@@ -176,11 +184,11 @@ _gnome_vfs_read_mime_from_buffer (GnomeVFSMimeSniffBuffer *buffer)
 	if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_EOF) {
 		return NULL;
 	}
-	G_LOCK (mime_mutex);
+	G_LOCK (gnome_vfs_mime_mutex);
 
 	mime_type = xdg_mime_get_mime_type_for_data (buffer->buffer, buffer->buffer_length);
 
-	G_UNLOCK (mime_mutex);
+	G_UNLOCK (gnome_vfs_mime_mutex);
 
 	return mime_type;
 	
@@ -190,15 +198,24 @@ const char *
 _gnome_vfs_get_mime_type_internal (GnomeVFSMimeSniffBuffer *buffer, const char *file_name, gboolean use_suffix)
 {
 	const char *result;
-	const char *zip_result;
+	const char *fn_result;
 
-	result = NULL;
+	result = fn_result = NULL;
+
+	/* Determine the mime type by filename as well, because
+	 * we need this for the subclassing check later
+	 */
+	if (file_name != NULL) {
+		fn_result = gnome_vfs_mime_type_from_name_or_default (file_name, NULL);
+	}
 
 	if (buffer != NULL) {
 		result = _gnome_vfs_read_mime_from_buffer (buffer);
-
+		
 		if (result != NULL && result != XDG_MIME_TYPE_UNKNOWN) {
+				
 			if ((strcmp (result, "application/x-ole-storage") == 0) ||
+			    (strcmp (result, "text/xml") == 0) ||
 			    (strcmp (result, "application/x-bzip") == 0) ||
 			    (strcmp (result, "application/x-gzip") == 0) ||
 			    (strcmp (result, "application/zip") == 0)) {
@@ -206,33 +223,36 @@ _gnome_vfs_get_mime_type_internal (GnomeVFSMimeSniffBuffer *buffer, const char *
 				 * that extensions are more reliable than magic
 				 * typing. If the file has a suffix, then use 
 				 * the type from the suffix.
-		 		 *
-				 * FIXME bugzilla.gnome.org 46867:
-				 * Allow specific mime types to override 
-				 * magic detection
-				 */
-				
-				if (file_name != NULL) {
-					zip_result = gnome_vfs_mime_type_from_name_or_default (file_name, NULL);
-					if (zip_result != NULL && zip_result != XDG_MIME_TYPE_UNKNOWN) {
-						return zip_result;
-					}
+		 		 */
+
+				if (fn_result != NULL && fn_result != XDG_MIME_TYPE_UNKNOWN) {
+					result = fn_result;
 				}
+				
+			} else if (fn_result && fn_result != XDG_MIME_TYPE_UNKNOWN) {
+				G_LOCK (gnome_vfs_mime_mutex);
+				if (xdg_mime_mime_type_subclass (fn_result, result)) {
+					result = fn_result;
+				}
+				G_UNLOCK (gnome_vfs_mime_mutex);
 			}
+			
 			return result;
 		}
 		
 		if (result == NULL || result == XDG_MIME_TYPE_UNKNOWN) {
 			if (_gnome_vfs_sniff_buffer_looks_like_text (buffer)) {
 				/* Text file -- treat extensions as a more 
-				 * accurate source of type information.
+				 * accurate source of type information BUT _only_ if
+				 * the extension is a subtype of text/plain.
 				 */
-				if (file_name != NULL) {
-					result = gnome_vfs_mime_type_from_name_or_default (file_name, NULL);
-				}
-	
-				if ((result != NULL) && (result != XDG_MIME_TYPE_UNKNOWN)) {
-					return result;
+				if ((fn_result != NULL) && (fn_result != XDG_MIME_TYPE_UNKNOWN)) {
+					G_LOCK (gnome_vfs_mime_mutex);
+					if (xdg_mime_mime_type_subclass (fn_result, "text/plain")) {
+						G_UNLOCK (gnome_vfs_mime_mutex);
+						return fn_result;
+					}
+					G_UNLOCK (gnome_vfs_mime_mutex);
 				}
 
 				/* Didn't find an extension match, assume plain text. */
@@ -243,12 +263,11 @@ _gnome_vfs_get_mime_type_internal (GnomeVFSMimeSniffBuffer *buffer, const char *
 			}
 		}
 	}
-	
+
 	if (use_suffix &&
-	    (result == NULL || result == XDG_MIME_TYPE_UNKNOWN) &&
-	    file_name != NULL) {
+	    (result == NULL || result == XDG_MIME_TYPE_UNKNOWN)) {
 		/* No type recognized -- fall back on extensions. */
-		result = gnome_vfs_mime_type_from_name_or_default (file_name, NULL);
+		result = fn_result;
 	}
 	
 	if (result == NULL) {
@@ -348,7 +367,7 @@ gnome_vfs_get_file_mime_type_internal (const char *path, const struct stat *opti
 	result = NULL;
 
 	/* get the stat info if needed */
-	if (optional_stat_info == NULL && stat (path, &tmp_stat_buffer) == 0) {
+	if (optional_stat_info == NULL && g_stat (path, &tmp_stat_buffer) == 0) {
 		optional_stat_info = &tmp_stat_buffer;
 	}
 
@@ -362,8 +381,10 @@ gnome_vfs_get_file_mime_type_internal (const char *path, const struct stat *opti
 			return "x-special/device-block";
 		} else if (S_ISFIFO(optional_stat_info->st_mode)) {
 			return "x-special/fifo";
+#ifdef S_ISSOCK
 		} else if (S_ISSOCK(optional_stat_info->st_mode)) {
 			return "x-special/socket";
+#endif
 		} else {
 			/* unknown entry type, return generic file type */
 			return GNOME_VFS_MIME_TYPE_UNKNOWN;
@@ -379,7 +400,7 @@ gnome_vfs_get_file_mime_type_internal (const char *path, const struct stat *opti
 	}
 	
 	if (!suffix_only) {
-		file = fopen(path, "r");
+		file = g_fopen(path, "r");
 	}
 
 	if (file != NULL) {
@@ -474,6 +495,8 @@ gnome_vfs_get_mime_type_from_uri (GnomeVFSURI *uri)
  * Tries to guess the mime type of the file @uri by
  * checking the file data using the magic patterns. Does not handle text files properly
  *
+ * Deprecated:
+ *
  * Returns the mime-type for this filename.
  */
 const char *
@@ -588,51 +611,12 @@ gboolean
 gnome_vfs_mime_type_is_equal (const char *a,
 			      const char *b)
 {
-	const gchar *alias_list;
-	gchar **aliases;
-	
 	g_return_val_if_fail (a != NULL, FALSE);
 	g_return_val_if_fail (b != NULL, FALSE);
 
-	/* First -- check if they're identical strings */
-	if (a == b)
-		return TRUE;
-	if (strcmp (a, b) == 0)
-		return TRUE;
-
-	/* next, check to see if 'a' is an alias for 'b' */
-	alias_list = gnome_vfs_mime_get_value (a, "aliases");
-	if (alias_list != NULL) {
-		int i;
-
-		aliases = g_strsplit (alias_list,
-				      ":",
-				      -1);
-		for (i = 0; aliases && aliases[i] != NULL; i++) {
-			if (strcmp (b, aliases[i]) == 0) {
-				g_strfreev (aliases);
-				return TRUE;
-			}
-		}
-		g_strfreev (aliases);
-	}
-
-	/* Finally, see if 'b' is an alias for 'a' */
-	alias_list = gnome_vfs_mime_get_value (b, "aliases");
-	if (alias_list != NULL) {
-		int i;
-
-		aliases = g_strsplit (alias_list,
-				      ":",
-				      -1);
-		for (i = 0; aliases && aliases[i] != NULL; i++) {
-			if (strcmp (a, aliases[i]) == 0) {
-				g_strfreev (aliases);
-				return TRUE;
-			}
-		}
-		g_strfreev (aliases);
-	}
+	G_LOCK (gnome_vfs_mime_mutex);
+	xdg_mime_mime_type_equal (a, b);
+	G_UNLOCK (gnome_vfs_mime_mutex);
 
 	return FALSE;
 }
@@ -661,56 +645,18 @@ GnomeVFSMimeEquivalence
 gnome_vfs_mime_type_get_equivalence (const char *mime_type,
 				     const char *base_mime_type)
 {
-	const gchar *parent_list;
-	char *supertype;
-
 	g_return_val_if_fail (mime_type != NULL, GNOME_VFS_MIME_UNRELATED);
 	g_return_val_if_fail (base_mime_type != NULL, GNOME_VFS_MIME_UNRELATED);
 
-	if (gnome_vfs_mime_type_is_equal (mime_type, base_mime_type))
+	if (gnome_vfs_mime_type_is_equal (mime_type, base_mime_type)) {
 		return GNOME_VFS_MIME_IDENTICAL;
-
-	supertype = gnome_vfs_get_supertype_from_mime_type (mime_type);
-
-	/* First, check if base_mime_type is a super type, and if it is, if
-	 * mime_type is one */
-	if (gnome_vfs_mime_type_is_supertype (base_mime_type)) {
-		if (! strcmp (supertype, base_mime_type)) {
-			g_free (supertype);
-			return GNOME_VFS_MIME_PARENT;
-		}
-	}
-
-	/* Both application/octet-stream and text/plain are special cases */
-	if (strcmp (base_mime_type, "text/plain") == 0 &&
-	    strcmp (supertype, "text/*") == 0) {
-		g_free (supertype);
-		return GNOME_VFS_MIME_PARENT;
-	}
-	g_free (supertype);
-
-	if (strcmp (base_mime_type, "application/octet-stream") == 0)
-		return GNOME_VFS_MIME_PARENT;
-
-	/* Then, check the parent-types of mime_type and compare  */
-	parent_list = gnome_vfs_mime_get_value (mime_type, "parent_classes");
-	if (parent_list) {
-		char **parents;
-		gboolean found_parent = FALSE;
-		int i;
-
-		parents = g_strsplit (parent_list, ":", -1);
-		for (i = 0; parents && parents[i] != NULL; i++) {
-			if (gnome_vfs_mime_type_get_equivalence (parents[i], base_mime_type)) {
-				found_parent = TRUE;
-				break;
-			}
-		}
-		g_strfreev (parents);
-
-		if (found_parent) {
+	} else {
+		G_LOCK (gnome_vfs_mime_mutex);
+		if (xdg_mime_mime_type_subclass (mime_type, base_mime_type)) {
+			G_UNLOCK (gnome_vfs_mime_mutex);
 			return GNOME_VFS_MIME_PARENT;
 		}
+		G_UNLOCK (gnome_vfs_mime_mutex);
 	}
 
 	return GNOME_VFS_MIME_UNRELATED;
@@ -722,7 +668,7 @@ static void
 file_date_record_update_mtime (FileDateRecord *record)
 {
 	struct stat s;
-	record->mtime = (stat (record->file_path, &s) != -1) ? s.st_mtime : 0;
+	record->mtime = (g_stat (record->file_path, &s) != -1) ? s.st_mtime : 0;
 }
 
 static FileDateRecord *
@@ -807,7 +753,7 @@ check_and_update_one (gpointer key, gpointer value, gpointer user_data)
 	record = (FileDateRecord *)value;
 	return_has_changed = (gboolean *)user_data;
 
-	if (stat (record->file_path, &s) != -1) {
+	if (g_stat (record->file_path, &s) != -1) {
 		if (s.st_mtime != record->mtime) {
 			record->mtime = s.st_mtime;
 			*return_has_changed = TRUE;
